#
# AAS mod for BF2: Special Forces conquest mode
#
# Version 1.0 (beta)
#
# The following copyright only applies to pieces of code added or changed
# in this file compared to the original gpm_cq.py included with BF2 1.02.
# Any other code is copyright of the copyright owner of that particular code.
#
# Copyright (c) 2005, MenDAKE <mendake_ddude@yahoo.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This is a modification of the original AAS mod for vanilla BF2 by 
# Jarkko Santala <santala@landot.org>, and was created and released by MenDAKE
# with permission.
#
# Please check SETTINGS below.
#

TAKEOVERTYPE_CAPTURE = 1
TAKEOVERTYPE_NEUTRALIZE = 2

SCORE_CAPTURE = 2
SCORE_NEUTRALIZE = 2
SCORE_CAPTUREASSIST = 1
SCORE_NEUTRALIZEASSIST = 1
SCORE_DEFEND = 1

Top = 0
Middle = 1
Bottom = 2

import host
import bf2
import math
from game.scoringCommon import addScore, RPL
from bf2 import g_debug

g_controlPoints = [] # cache, as this map won't change

global aasTip
global aasLast
global aasNextCp
global aasRadius
global aasVehicleCap
global aasCpBalance
global aasDebug
global aasL10n
global aasL10nCp
global aasL10nText

###############################################################################
#
# landot AAS mod - SETTINGS
#
#
# The CP area sizes will be calculated using the following formula:
#
# Original CP Radius * ( Max Number Of Players / aasRadius )
#
# If you set the aasRadius to 0 the CP area sizes will stay normal.
# With default setting 10:  10 players = the same, 30 players = 3 times bigger
#
aasRadius = 10
#
# Choose localization by setting the variable aasL10n
#
# 0 = English
# 1 = French
# 2 = German
#
aasL10n = 0
#
# Enable/disable Commander
#
# 1 = Enable
# 0 = Disable
#
aasCommander = 0
#
# Using the following setting you can choose whether CPs per team will
# be balanced at the start of each map. Enabling this setting means that
# the battle will always start in the middle of the map.
#
# True  = Balance number of CPs per team at map start
# False = Don't balance, use default CP ownerships
#
aasCpBalance = True
#
# No user serviceable parts below this point!
#
###############################################################################

aasDebug = 0
aasLast = [ 0, 0, 0 ]
aasNextCp = [ 0, 0, 0 ]

aasTip = "|C     * AAS * |C    "

aasL10nCp = { "asylum" : [ "Asylum", " ", " " ], 
		"entrance" : [ "Entrance", " ", " " ],
		"church" : [ "Church", " ", " " ], 
		"uppervillage" : [ "Upper Village", " ", " " ] }

aasL10nTexts = [ [ "Flags must be captured in correct order!", "You have entered the ", " must be captured before the ", "Proceed to the " ] ,
		 [ "Capturez les points dans le bon ordre!", "Vous tes au point : ", " : doit tre captur avant : ", "Capturez : " ] ,
		 [ "Die Flaggen mssen in einer bestimmten Reihenfolge erobert werden!", "Gerade erobert: ", " muss zuerst erobert werden vor ", "Nchster Eroberungspunkt: " ] ]

aasL10nText = aasL10nTexts[aasL10n]

def aasSetNextCp(aasTeam, aasInit):
	
	aasMaps = { "16" :
		    { "Devils_Perch" :
		      [ "cp_16p_dp_asylum", "cp_16p_dp_gated_entrance", "cp_16p_dp_church", "cp_16p_dp_uppervillage" ] ,
		      "Ghost_Town" :
		      [ "cp_16p_gt_suburbs", "cp_16p_gt_centralsquare", "cp_16p_gt_convoy", "cp_16p_gt_waterstation" ] ,
		      "Iron_Gator" :
		      [ "cp_16p_ms_engine_room", "cp_16p_ms_bunks", "cp_16p_ms_hanger_bay_fore", "cp_16p_ms_mess_hall" ] ,
		      "Leviathan" :
		      [ "cp_16p_lev_beach", "cp_16p_lev_powerstation", "cp_16p_lev_supplypier", "cp_16p_lev_commbuilding" ] ,
		      "Mass_Destruction" :
		      [ "cp_16p_md_shore", "cp_16p_md_facility", "cp_16p_md_yard", "cp_16p_md_scrapyard" ] ,
		      "Night_Flight" :
		      [ "cp_16p_nf_cargodepot", "cp_16p_nf_cargoplane", "cp_16p_nf_terminalcomplex" ] ,
		      "Surge" :
		      [ "cp_16p_surge_commandbunker", "cp_16p_surge_launchpad", "cp_16p_surge_baseentrance" ] ,
		      "Warlord" :
		      [ "cp_16p_wl_palacegardens", "cp_16p_wl_vault", "cp_16p_wl_conferenceroom", "cp_16p_wl_securityoffice" ] } ,
		    "32" :
		    { "Devils_Perch" :
		      [ "cp_32p_dp_asylum", "cp_32p_dp_church", "cp_32p_dp_hilltop", "cp_32p_dp_uptown", "cp_32p_dp_lowtown" ] ,
		      "Ghost_Town" :
		      [ "cp_32p_gt_waterfront", "cp_32p_gt_uppernorthside", "cp_32p_gt_uppersouthside", "cp_32p_gt_convoy", "cp_32p_gt_waterstation" ] ,
		      "Iron_Gator" :
		      [ "cp_32p_ms_controltower", "cp_32p_ms_welldeck", "cp_32p_ms_hanger_bay_aft", "cp_32p_ms_controltowerexterior", "cp_32p_ms_mess_hall", "cp_32p_ms_hanger_bay_fore" ] , 
		      "Leviathan" :
		      [ "cp_32p_lev_fueldepot", "cp_32p_lev_supplydepot", "cp_32p_lev_supplypier_south", "cp_32p_lev_subpenleft", "cp_32p_lev_subpenright", "cp_32p_lev_armory"] ,
		      "Mass_Destruction" :
		      [ "cp_32p_md_scrapyard", "cp_32p_md_coolingstation", "cp_32p_md_yard", "cp_32p_md_facility", "cp_32p_md_shore" ] ,
		      "Night_Flight" :
		      [ "cp_32p_nf_aircontroltower", "cp_32p_nf_cargodepot", "cp_32p_nf_cargoplane", "cp_32p_nf_terminalcomplex" ] ,
		      "Surge" :
		      [ "cp_32p_surge_supplydepot", "cp_32p_surge_commandbunker", "cp_32p_surge_launchpad", "cp_32p_surge_baseentrance", "cp_32p_surge_innerridge" ] ,
		      "Warlord" :
		      [ "cp_32p_wl_vault", "cp_32p_wl_conferenceroom", "cp_32p_wl_stationplaza", "cp_32p_wl_securityoffice", "cp_32p_wl_insurgenthideout", "cp_32p_wl_stationplaza",  "cp_32p_wl_parkinggarage"] } ,
		     "64" :
		    { "Devils_Perch" :
		      [ "cp_64p_dp_waterfront", "cp_64p_dp_church", "cp_64p_dp_hilltop", "cp_64p_dp_uptown", "cp_64p_dp_lowtown", "cp_64p_dp_powerstation", "cp_64p_dp_causeway", "cp_64p_dp_marketsquare", "cp_64p_dp_portfactory" ] ,
		      "Ghost_Town" :
		      [ "cp_64p_gt_waterfront", "cp_64p_gt_uppernorthside", "cp_64p_gt_uppersouthside", "cp_64p_gt_convoy", "cp_64p_gt_waterstation", "cp_64p_gt_beachstreet", "cp_64p_gt_oilplatform", "cp_64p_gt_oilfield" ] ,
		      "Iron_Gator" :
		      [ "cp_64p_ms_controltower", "cp_64p_ms_welldeck", "cp_64p_ms_hanger_bay_aft", "cp_64p_ms_controltowerexterior", "cp_64p_ms_mess_hall", "cp_64p_ms_hanger_bay_fore" ] ,
		      "Leviathan" :
		      [ "cp_64p_lev_beach", "cp_64p_lev_powerstation", "cp_64p_lev_commbuilding", "cp_64p_lev_subpen02", "cp_64p_lev_subpen01", "cp_642p_lev_armory",  "cp_64p_lev_supply_depot", "cp_64p_lev_fuel_depot"] ,
		      "Mass_Destruction" :
		      [ "cp_64p_md_island", "cp_64p_md_scrapyard", "cp_64p_md_coolingstation", "cp_64p_md_yard", "cp_64p_md_facility", "cp_64p_md_shore" ] ,
		      "Night_Flight" :
		      [ "cp_64p_nf_aircontroltower", "cp_64p_nf_cargodepot", "cp_64p_nf_cargpolane", "cp_64p_nf_terminalcomplex", "cp_64p_nf_helipad" ] ,
		      "Surge" :
		      [ "cp_64p_surge_commandbunker", "cp_64p_surge_fueldepot", "cp_64p_surge_launchpad", "cp_64p_surge_supplydepot", "cp_64p_surge_innerridge", "cp_64p_surge_outerridge", "cp_64p_surge_checkpoint", "cp_64p_surge_outerridge" ] ,
		      "Warlord" :
		      [ "cp_64p_wl_vault", "cp_64p_wl_conferenceroom", "cp_64p_wl_securityoffice", "cp_64p_wl_mosque", "cp_64p_wl_insurgenthideout", "cp_64p_wl_stationplaza", "cp_64p_wl_parkinggarage" ] } }
		    
	aasCps = False
	for obj in bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint'):
		obj = obj.templateName
		aasMapSize = obj[3:5]
		break
	aasMapsOfSize = aasMaps[aasMapSize]
	aasCps = aasMapsOfSize[bf2.gameLogic.getMapName()]

	aasTeam2 = 2
	if aasCps:
		if aasTeam == 2:
			aasTeam2 = 1
			aasCps.reverse()
		for aasCnt in range(0,len(aasCps)):
			aasNumOfCps = 0
			aasTeamCps = 0
			for obj in bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint'):
				aasNumOfCps += 1
				if obj.cp_getParam('team') == aasTeam:
					aasTeamCps += 1
				if obj.templateName.lower() == aasCps[aasCnt]:
					aasCpId = obj
				elif obj.templateName.lower() == aasCps[aasCnt-1]:
					aasPrevCpId = obj
			if aasInit:
				if aasTeamCps < int(aasNumOfCps / 2):
					aasCpId.cp_setParam('flag', aasTeam)
					aasCpId.cp_setParam('team', aasTeam)
					aasCpId.cp_setParam('takeOverChangePerSecond', 100)
			elif aasCpId.cp_getParam('team') != aasTeam:
				if aasCnt == 0 or aasPrevCpId.cp_getParam('team') == aasTeam:
					aasNextCp[aasTeam] = aasCps[aasCnt]
					break
					
def aasCpName(aasCp):
	aasCp = aasCp.lower()
	aasCp = aasL10nCp[aasCp[10:]]
	aasCp = aasCp[aasL10n]
	return aasCp


def init():
	# events hook
	host.registerGameStatusHandler(onGameStatusChanged)
	if host.sgl_getIsAIGame() == 1:
		host.sh_setEnableCommander(aasCommander)
	else:
		host.sh_setEnableCommander(aasCommander)
		
	host.registerHandler('TimeLimitReached', onTimeLimitReached, 1)	

	if g_debug: print "gpm_cq.py initialized"
		
		
		
def deinit():
	bf2.triggerManager.destroyAllTriggers()
	global g_controlPoints
	g_controlPoints = []
	host.unregisterGameStatusHandler(onGameStatusChanged)
	if g_debug: print "gpm_cq.py uninitialized"



def onGameStatusChanged(status):

	global g_controlPoints
	if status == bf2.GameStatus.Playing: 

		if aasCpBalance:
			aasSetNextCp(1, 1)
			aasSetNextCp(2, 1)

		# add control point triggers
		g_controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
		for obj in g_controlPoints:
			radius = float(obj.getTemplateProperty('radius'))
			isHemi = int(obj.cp_getParam('isHemisphere'))

			if aasRadius > 0:
				aasRadiusFactor = float(host.ss_getParam('maxPlayers')) / aasRadius
				if aasRadiusFactor > 1:
					radius = round(radius * aasRadiusFactor)
			
			if isHemi != 0:
				id = bf2.triggerManager.createHemiSphericalTrigger(obj, onCPTrigger, '<<PCO>>', radius, (1, 2, 3))
			else:
				id = bf2.triggerManager.createRadiusTrigger(obj, onCPTrigger, '<<PCO>>', radius, (1, 2, 3))			
			obj.triggerId = id
			obj.lastAttackingTeam = 0
			if obj.cp_getParam('team') > 0:
				obj.flagPosition = Top
			else:
				obj.flagPosition = Bottom

		host.registerHandler('ControlPointChangedOwner', onCPStatusChange)

		# setup ticket system
		ticketsTeam1 = calcStartTickets(bf2.gameLogic.getDefaultTickets(1))
		ticketsTeam2 = calcStartTickets(bf2.gameLogic.getDefaultTickets(2))
		
		bf2.gameLogic.setTickets(1, ticketsTeam1)
		bf2.gameLogic.setTickets(2, ticketsTeam2)

		bf2.gameLogic.setTicketState(1, 0)
		bf2.gameLogic.setTicketState(2, 0)

		bf2.gameLogic.setTicketLimit(1, 1, 0)
		bf2.gameLogic.setTicketLimit(2, 1, 0)
		bf2.gameLogic.setTicketLimit(1, 2, 10)
		bf2.gameLogic.setTicketLimit(2, 2, 10)
		bf2.gameLogic.setTicketLimit(1, 3, int(ticketsTeam1*0.1))
		bf2.gameLogic.setTicketLimit(2, 3, int(ticketsTeam2*0.1))
		bf2.gameLogic.setTicketLimit(1, 4, int(ticketsTeam1*0.2))
		bf2.gameLogic.setTicketLimit(2, 4, int(ticketsTeam1*0.2))
		
		host.registerHandler('TicketLimitReached', onTicketLimitReached)
		updateTicketLoss()

		# player events
		host.registerHandler('PlayerDeath', onPlayerDeathCQ)
		host.registerHandler('PlayerKilled', onPlayerKilledCQ)
		host.registerHandler('PlayerRevived', onPlayerRevived)
		host.registerHandler('PlayerSpawn', onPlayerSpawn)
		host.registerHandler('EnterVehicle', onEnterVehicle)
		host.registerHandler('ExitVehicle', onExitVehicle)

		if g_debug: print "Conquest gamemode initialized."
	else:
		bf2.triggerManager.destroyAllTriggers()
		g_controlPoints = []



def calcStartTickets(mapDefaultTickets):
	return int(mapDefaultTickets * (bf2.serverSettings.getTicketRatio() / 100.0))
	
	
	
def onTimeLimitReached(value):
	team1tickets = bf2.gameLogic.getTickets(1)
	team2tickets = bf2.gameLogic.getTickets(2)
	
	winner = 0
	victoryType = 0
	if team1tickets > team2tickets:
		winner = 1
		victoryType = 3
	elif team2tickets > team1tickets:
		winner = 2
		victoryType = 3
	

	host.sgl_endGame(winner, victoryType)



# update ticket system
def updateTicketLoss():
	areaValueTeam1 = 0
	areaValueTeam2 = 0
	totalAreaValue = 0
	numCpsTeam0 = 0
	numCpsTeam1 = 0
	numCpsTeam2 = 0
	
	# calculate control point area value for each team
	for obj in g_controlPoints:
		team = obj.cp_getParam('team')
		if team == 1:
			areaValueTeam1 += obj.cp_getParam('areaValue', team)
			totalAreaValue += areaValueTeam1
			numCpsTeam1 += 1
		elif team == 2:
			areaValueTeam2 += obj.cp_getParam('areaValue', team)
			totalAreaValue += areaValueTeam2
			numCpsTeam2 += 1
		else:
			numCpsTeam0 += 1
			totalAreaValue += 0
		
	# check if a team has no control points
	if numCpsTeam1 == 0 or numCpsTeam2 == 0:
		if numCpsTeam1 == 0:
			losingTeam = 1
			winningTeam = 2
		else:
			losingTeam = 2
			winningTeam = 1
		
		# check if there is anyone alive
		foundLivingPlayer = False
		for p in bf2.playerManager.getPlayers():
			if p.getTeam() == losingTeam and p.isAlive():
				foundLivingPlayer = True
				break
				
		if not foundLivingPlayer:

			# drop tickets
			ticketLossPerSecond = bf2.gameLogic.getDefaultTicketLossAtEndPerMin()
			bf2.gameLogic.setTicketChangePerSecond(losingTeam, -ticketLossPerSecond)
			bf2.gameLogic.setTicketChangePerSecond(winningTeam, 0)
			
			return

	
	# update ticket loss
	team1AreaOverweight = areaValueTeam1 - areaValueTeam2
	percentualOverweight = 1.0
	if totalAreaValue != 0:
		percentualOverweight = abs(team1AreaOverweight / totalAreaValue)
	
	ticketLossPerSecTeam1 = calcTicketLossForTeam(1, areaValueTeam2, -team1AreaOverweight)
	ticketLossPerSecTeam2 = calcTicketLossForTeam(2, areaValueTeam1, team1AreaOverweight)
	bf2.gameLogic.setTicketChangePerSecond(1, -ticketLossPerSecTeam1)
	bf2.gameLogic.setTicketChangePerSecond(2, -ticketLossPerSecTeam2)


	
# actual ticket loss calculation function
def calcTicketLossForTeam(team, otherTeamAreaValue, otherTeamAreaOverweight):
	if otherTeamAreaValue >= 100 and otherTeamAreaOverweight > 0:
		ticketLossPerSecond = (bf2.gameLogic.getDefaultTicketLossPerMin(team) / 60.0) * (otherTeamAreaOverweight / 100.0)
		return ticketLossPerSecond
	else:
		return 0


	
DOWNWARDS = -1
UPWARDS = 1	

# called when tickets reach a predetermined limit (negativ value means that the tickets have become less than the limit)
def onTicketLimitReached(team, limitId):
	if (limitId == -1):
		if (team == 1):
			winner = 2
		
		elif (team == 2):
			winner = 1
		
		bf2.gameLogic.setTicketState(1, 0)
		bf2.gameLogic.setTicketState(2, 0)
	
		host.sgl_endGame(winner, 3)		

	# update ticket state
	else:
		updateTicketWarning(team, limitId)

	

# called when the ticket state should be updated (for triggering messages and sounds based on tickets left)
def updateTicketWarning(team, limitId):

	oldTicketState = bf2.gameLogic.getTicketState(team)
	newTicketState = 0
	
	if (oldTicketState >= 10):
		newTicketState = 10		

	if (limitId == -2):
		newTicketState = 10
	
	elif (limitId == 2):
		newTicketState = 0		

	elif (limitId == -3):
		newTicketState += 2

	elif (limitId == -4):
		newTicketState += 1

	if (oldTicketState != newTicketState):
		bf2.gameLogic.setTicketState(team, newTicketState)
		
	
	
# called when someone enters or exits cp radius
def onCPTrigger(triggerId, cp, vehicle, enter, userData):
	if not cp.isValid(): return

	#host.rcon_invoke('game.sayall ' + ' "' + cp.templateName + '"')

	# can this cp be captured at all?
	if cp.cp_getParam('unableToChangeTeam') != 0:
		return					
		
	if enter:
		for p in vehicle.getOccupyingPlayers():
			cp = getOccupyingCP(p)
			if cp != None:
				if not p.getIsInsideCP():
					if g_debug: print "Resetting enterPctAt for player ", p.getName()
					p.enterCpAt = host.timer_getWallTime()
	
	if vehicle:	
		for p in vehicle.getOccupyingPlayers():
			p.setIsInsideCP(enter)			

	# count people in radius
	team1Occupants = 0
	team2Occupants = 0

	pcos = bf2.triggerManager.getObjects(cp.triggerId)
	for o in pcos:
		if not o: continue # you can get None in the result tuple when the host can't figure out what object left the trigger
		if o.getParent(): continue # getOccupyingPlayers returns all players downwards in the hierarchy, so dont count them twice
		for p in o.getOccupyingPlayers():
			if p.isAlive() and not p.isManDown():
				
				if not p.killed:
					if p.getTeam() == 1:
						team1Occupants += 1
					elif p.getTeam() == 2:
						team2Occupants += 1


	# determine who is taking control
	team1OverWeight = team1Occupants - team2Occupants
	attackOverWeight = 0

	if team1OverWeight > 0:
		attackingTeam = 1
	elif team1OverWeight < 0:
		attackingTeam = 2
	else:
		attackingTeam = 0

	
	if team1Occupants == 0 and team2Occupants == 0:

		# nobody here, slowly go back to owning team
		if cp.cp_getParam('team') == 0:
			attackOverWeight = -0.5
		else:
			attackOverWeight = 0.5
			
		timeToChangeControl = cp.cp_getParam('timeToLoseControl')

	else:

		aasTeam = attackingTeam

		aasSetNextCp(aasTeam, 0)

		if cp.cp_getParam('team') != aasTeam:
			if aasNextCp[aasTeam] != cp.templateName.lower():
				aasMsg = 'game.sayTeam ' + str(aasTeam) + ' "' + aasTip + aasCpName(aasNextCp[aasTeam]) + aasL10nText[2] + aasCpName(cp.templateName) + aasTip + '"'
				if aasMsg != aasLast[aasTeam]:
					aasLast[aasTeam] = aasMsg
					host.rcon_invoke(aasMsg)
				return

		# raise flag if already ours, or at bottom and neutral. Otherwise lower first.
		if cp.cp_getParam('flag') == attackingTeam or (cp.flagPosition == Bottom and cp.cp_getParam('team') == 0):

			# our flag, raise
			attackOverWeight = abs(team1OverWeight)
			timeToChangeControl = cp.cp_getParam('timeToGetControl')
		else:

			# other team raised flag, lower first
			attackOverWeight = - abs(team1OverWeight)
			timeToChangeControl = cp.cp_getParam('timeToLoseControl')


	if cp.cp_getParam('onlyTakeableByTeam') != 0 and cp.cp_getParam('onlyTakeableByTeam') != attackingTeam:
		return


	# flag can only be changed when at bottom
	if cp.flagPosition == Bottom:
		cp.cp_setParam('flag', attackingTeam)


	# calculate flag raising/lowering speed
	if timeToChangeControl > 0:
		takeOverChangePerSecond = 1.0 * attackOverWeight / timeToChangeControl
	else:
		takeOverChangePerSecond = 0.0

	if (cp.flagPosition == Top and takeOverChangePerSecond > 0) or (cp.flagPosition == Bottom and takeOverChangePerSecond < 0):
		takeOverChangePerSecond = 0.0

	if abs(takeOverChangePerSecond) > 0:
		cp.flagPosition = Middle
				
	cp.cp_setParam('takeOverChangePerSecond', takeOverChangePerSecond)

			
	
# called when a control point flag reached top or bottom
def onCPStatusChange(cp, top):

	playerId = -1
	takeoverType = -1
	newTeam = -1
	scoringTeam = -1
	
	if top:	cp.flagPosition = Top
	else:   cp.flagPosition = Bottom
	
	# determine capture / neutralize / defend
	if cp.cp_getParam('team') != 0:

		if top:
			# regained flag, do nothing
			pass
			
		else:
			# neutralize
			newTeam = 0
			if cp.cp_getParam('team') == 1:
				scoringTeam = 2
			else:
				scoringTeam = 1
				
			takeoverType = TAKEOVERTYPE_NEUTRALIZE

	else:

		if top:
			# capture
			newTeam = cp.cp_getParam('flag')
			scoringTeam = newTeam
			takeoverType = TAKEOVERTYPE_CAPTURE

		else:
			# hit bottom, but still neutral
			pass

	
	# scoring
	if takeoverType > 0:
		pcos = bf2.triggerManager.getObjects(cp.triggerId)
	
		# count number of players
		scoringPlayers = []
		firstPlayers = []
		for o in pcos:
			for p in o.getOccupyingPlayers():
				if p.isAlive() and not p.isManDown() and p.getTeam() == scoringTeam:
					if len(firstPlayers) == 0 or p.enterCpAt < firstPlayers[0].enterCpAt:
						firstPlayers = [p]
					elif p.enterCpAt == firstPlayers[0].enterCpAt:
						firstPlayers += [p]
					
					if not p in scoringPlayers:
						scoringPlayers += [p]
	
		# deal score
		for p in scoringPlayers:
			oldScore = p.score.score;
			if takeoverType == TAKEOVERTYPE_CAPTURE:
				if p in firstPlayers:
					p.score.cpCaptures += 1
					addScore(p, SCORE_CAPTURE, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 0) #12 = Conquest, 0 = Capture
					playerId = p.index
				else:
					p.score.cpAssists += 1
					addScore(p, SCORE_CAPTUREASSIST, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 2) #12 = Conquest, 2 = Assist


			elif takeoverType == TAKEOVERTYPE_NEUTRALIZE:
				if p in firstPlayers:
					p.score.cpNeutralizes += 1
					addScore(p, SCORE_NEUTRALIZE, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 3) #12 = Conquest, 3 = Neutralize
				else:
					p.score.cpNeutralizeAssists += 1
					addScore(p, SCORE_NEUTRALIZEASSIST, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 4) #12 = Conquest, 4 = Neutralize assist
					
				

	# immediate ticket loss for opposite team
	enemyTicketLossInstant = cp.cp_getParam('enemyTicketLossWhenCaptured')
	if enemyTicketLossInstant > 0 and newTeam > 0:
		
		if newTeam == 1:
			punishedTeam = 2
		elif newTeam == 2:
			punishedTeam = 1
		
		tickets = bf2.gameLogic.getTickets(punishedTeam)
		tickets -= enemyTicketLossInstant
		bf2.gameLogic.setTickets(punishedTeam, tickets)
	
	
	# update control point	
	cp.cp_setParam('playerId', playerId) #always set player first
	if newTeam != -1 and cp.cp_getParam('team') != newTeam:
		cp.cp_setParam('team', newTeam)
	onCPTrigger(cp.triggerId, cp, 0, 0, 0)
	updateTicketLoss()

	for aasTeam in [ 1, 2 ]:
		aasSetNextCp(aasTeam, 0)
		aasMsg = 'game.sayTeam ' + str(aasTeam) + ' "' + aasTip + aasL10nText[3] + aasCpName(aasNextCp[aasTeam]) + aasTip + '"'
		if aasMsg != aasLast[aasTeam]:
			aasLast[aasTeam] = aasMsg
			host.rcon_invoke(aasMsg)
				
				
def onPlayerDeathCQ(victim, vehicle):		

	# punish team with one ticket
	if victim != None:
		team = victim.getTeam()
		teamTickets = bf2.gameLogic.getTickets(team)
		teamTickets -= 1
		bf2.gameLogic.setTickets(team, teamTickets)

	# check if it was the last player
	foundLivingPlayer = False
	for p in bf2.playerManager.getPlayers():
		if p != victim and p.getTeam() == victim.getTeam() and p.isAlive():
			foundLivingPlayer = True
	
	if not foundLivingPlayer:
		updateTicketLoss()
	


def onPlayerKilledCQ(victim, attacker, weapon, assists, object):
	if not victim: 
		return
	
	victim.killed = True
	
	# update flag takeover status if victim was in a CP radius
	cp = getOccupyingCP(victim)
	if cp != None:
		onCPTrigger(-1, cp, victim.getVehicle(), False, None)

		# give defend score if killing enemy within cp radius
		if attacker != None and attacker.getTeam() != victim.getTeam()\
		   and cp.cp_getParam('unableToChangeTeam') == 0 and cp.cp_getParam('onlyTakeableByTeam') == 0:
		
			if cp != None and cp.cp_getParam('team') == attacker.getTeam():
				attacker.score.cpDefends += 1
				addScore(attacker, SCORE_DEFEND, RPL)
				bf2.gameLogic.sendGameEvent(attacker, 12, 1) #12 = Conquest, 1 = Defend
			
			
			
def onPlayerRevived(victim, attacker):

	# update flag takeover status if victim was in a CP radius
	victim.killed = False
	
	cp = getOccupyingCP(victim)
	if cp != None:
		onCPTrigger(-1, cp, victim.getVehicle(), True, None)
	
	
	
def onPlayerSpawn(player, soldier):
	player.killed = False
	
	aasTeam = player.getTeam()

	aasMsg = 'game.sayTeam ' + str(aasTeam) + ' "' + aasTip + aasL10nText[0] + aasTip + '"'
	if aasMsg != aasLast[aasTeam]:
		aasLast[aasTeam] = aasMsg
		host.rcon_invoke(aasMsg)

	aasSetNextCp(aasTeam, 0)

	aasMsg = 'game.sayTeam ' + str(aasTeam) + ' "' + aasTip + aasL10nText[3] + aasCpName(aasNextCp[aasTeam]) + aasTip + '"'
	if aasMsg != aasLast[aasTeam]:
		aasLast[aasTeam] = aasMsg
		host.rcon_invoke(aasMsg)
	
	
def onEnterVehicle(player, vehicle, freeSoldier = False):

	# update flag takeover status if player in a CP radius
	cp = getOccupyingCP(player)
	if cp != None:
		player.setIsInsideCP(True)
	else:		
		player.setIsInsideCP(False)



def onExitVehicle(player, vehicle):

	# update flag takeover status if player in a CP radius
	cp = getOccupyingCP(player)
	if cp != None:
		player.setIsInsideCP(True)
	else:		
		player.setIsInsideCP(False)
		
		
		
# find cp that player is occupying, if any		
def getOccupyingCP(player):
	vehicle = player.getVehicle()
	playerPos = vehicle.getPosition()
	
	# find closest CP
	closestCP = None
	if len(g_controlPoints) == 0: return None
	for obj in g_controlPoints:
		distanceTo = getVectorDistance(playerPos, obj.getPosition())
		if closestCP == None or distanceTo < closestCPdist:
			closestCP = obj
			closestCPdist = distanceTo
	
	# is the player in radius?
	pcos = bf2.triggerManager.getObjects(closestCP.triggerId)
	for o in pcos:
		if o == player.getDefaultVehicle():
			# Player is DEFAULT vehicle - this is needed when called from onEnterVehicle
			return closestCP
		else:
			for p in o.getOccupyingPlayers():
				if p == player:
					return closestCP
	
	return None



# get distance between two positions
def getVectorDistance(pos1, pos2):
	diffVec = [0.0, 0.0, 0.0]
	diffVec[0] = math.fabs(pos1[0] - pos2[0])
	diffVec[1] = math.fabs(pos1[1] - pos2[1])
	diffVec[2] = math.fabs(pos1[2] - pos2[2])
	 
	return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])
		
		
	
